class EventSlot:
    def __init__(self):
        self.targets = []

    def __iadd__(self, f):
        self.targets.append(f)
        return self

    def __isub__(self, f):
        while f in self.targets:
            self.targets.remove(f)
        return self

    def __len__(self):
        return len(self.targets)

    def __iter__(self):
        def gen():
            for target in self.targets:
                yield target
        return gen()

    def __getitem__(self, key):
        return self.targets[key]



from enum import IntEnum
from sys import float_info
from typing import Protocol, List, Union, Dict, Any, Optional

from typing_extensions import Self

from .training_protocols import PredicateContext, TrainingPhaseProtocol, class_from_dict, class_to_dict

_FLOAT_COMPARISON = float_info.epsilon * 2


class TrainingPredicate(Protocol):
    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        return False

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return None

    def to_dict(self) -> Dict[str, Any]:
        return class_to_dict(self.__class__.__name__, self.__class__.__module__, self._to_dict())

    @classmethod
    def from_dict(cls, data: Optional[Dict[str, Any]]):
        return class_from_dict(data)


class CompoundPredicate(TrainingPredicate):
    class CompoundType(IntEnum):
        AND = 0
        OR = 1

    def __init__(self, compound_type: CompoundType, predicates: List[TrainingPredicate]):
        self.compound_type = compound_type
        self.predicates = predicates

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        if self.compound_type == CompoundPredicate.CompoundType.AND:
            return all(predicate.evaluate(phase, context) for predicate in self.predicates)
        else:
            return any(predicate.evaluate(phase, context) for predicate in self.predicates)

    def _to_dict(self) -> Dict[str, Any]:
        return {
            "compound_type": self.compound_type.name,
            "predicates": [predicate.to_dict() for predicate in self.predicates]
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        compound_type = CompoundPredicate.CompoundType[data["compound_type"]]
        predicates = [TrainingPredicate.from_dict(d) for d in data["predicates"]]

        return CompoundPredicate(compound_type, predicates)


class NumberComparisonPredicate(TrainingPredicate):
    class NumberComparisonType(IntEnum):
        LT = 0
        LTE = 1
        EQ = 2
        GTE = 3
        GT = 4

    def __init__(self, comparison_type: NumberComparisonType, path: str, value: Union[int, float]):
        self.comparison_type = comparison_type
        if not isinstance(path, list):
            self.path = path.split(".")
        else:
            self.path = path
        self.value = value

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        eval_obj = context
        for part in self.path:
            if eval_obj and part:
                if hasattr(eval_obj, part):
                    eval_obj = eval_obj.__getattribute__(part)
                else:
                    eval_obj = None
                    break

        if eval_obj is None:
            return False

        if self.comparison_type == NumberComparisonPredicate.NumberComparisonType.LT:
            return eval_obj < self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.LTE:
            return eval_obj <= self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.EQ:
            if isinstance(eval_obj, float) or isinstance(self.value, float):
                return abs(eval_obj - self.value) < _FLOAT_COMPARISON
            else:
                return eval_obj == self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.GTE:
            return eval_obj >= self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.GT:
            return eval_obj > self.value

        return False

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return {
            "comparison_type": self.comparison_type.name,
            "path": self.path,
            "value": self.value
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        comparison_type = NumberComparisonPredicate.NumberComparisonType[data["comparison_type"]]
        path = data["path"]
        value = data["value"]

        return NumberComparisonPredicate(comparison_type, path, value)


class PythonPredicate(TrainingPredicate):
    def __init__(self, python_code: str):
        self.python_code = python_code

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        pass

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return {
            "python_code": self.python_code
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]):
        python_code = data["python_code"]
        return PythonPredicate(python_code)



import math
import uuid
from enum import IntEnum
from typing import List, Dict, Any, Optional, Callable
from typing_extensions import Self

import humps

from .event_slot import EventSlot
from .training_protocols import PredicateContext, TrainingPhaseProtocol, TrainingActionProtocol
from .training_predicate import TrainingPredicate
from .training_action import TrainingAction
from .training_progress import TrainingProgress


class SessionResult(IntEnum):
    NONE = 0
    FALLBACK = 10
    ADVANCE = 20


class TrainingPhase(TrainingPhaseProtocol):
    def __init__(self, phase_id: Optional[str] = None):
        self._phase_id = phase_id or str(uuid.uuid4())
        self._name: str = ""
        self._description: str = ""
        self._fallback_condition: Optional[TrainingPredicate] = None
        self._advance_condition: Optional[TrainingPredicate] = None
        self._session_actions: List[TrainingActionProtocol] = []

        self._is_pellet_delivery_enabled: bool = False
        self._is_pellet_cover_enabled: bool = False
        self._starting_baseline_intensity: int = 0
        self._pellet_hands_min_distance: float = 0.0
        self._is_pellet_shift_enabled: bool = False
        self._is_auto_clamp_enabled: bool = False
        self._auto_clamp_no_activity_release_delay: float = 0.0
        self._auto_clamp_release_load_count: int = 0

        self._progress = TrainingProgress()

        self._would_advance: bool = False
        self._would_fallback: bool = False

        self.property_changed = EventSlot()

    @property
    def phase_id(self) -> str:
        return self._phase_id

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, value: str) -> None:
        self._description = value

    @property
    def fallback_predicate(self) -> Optional[TrainingPredicate]:
        return self._fallback_condition

    @fallback_predicate.setter
    def fallback_predicate(self, value: Optional[TrainingPredicate]) -> None:
        self._fallback_condition = value

    @property
    def advance_predicate(self) -> Optional[TrainingPredicate]:
        return self._advance_condition

    @advance_predicate.setter
    def advance_predicate(self, value: Optional[TrainingPredicate]) -> None:
        self._advance_condition = value

    @property
    def session_actions(self) -> List[TrainingActionProtocol]:
        return self._session_actions

    @session_actions.setter
    def session_actions(self, value: List[Callable[[Self, PredicateContext], None]]) -> None:
        self._session_actions = value

    @property
    def progress(self) -> TrainingProgress:
        return self._progress

    @progress.setter
    def progress(self, value: TrainingProgress) -> None:
        self._progress = self._on_property_changed("progress", value, self._progress)

    @property
    def is_pellet_delivery_enabled(self) -> bool:
        return self._is_pellet_delivery_enabled

    @is_pellet_delivery_enabled.setter
    def is_pellet_delivery_enabled(self, value: bool) -> None:
        self._is_pellet_delivery_enabled = value

    @property
    def is_pellet_cover_enabled(self) -> bool:
        return self._is_pellet_cover_enabled

    @is_pellet_cover_enabled.setter
    def is_pellet_cover_enabled(self, value: bool) -> None:
        self._is_pellet_cover_enabled = value

    @property
    def starting_baseline_intensity(self) -> int:
        return self._starting_baseline_intensity

    @starting_baseline_intensity.setter
    def starting_baseline_intensity(self, value: int) -> None:
        self._starting_baseline_intensity = value

    @property
    def pellet_hands_min_distance(self) -> float:
        return self._pellet_hands_min_distance

    @pellet_hands_min_distance.setter
    def pellet_hands_min_distance(self, value: float) -> None:
        self._pellet_hands_min_distance = value

    @property
    def is_pellet_shift_enabled(self) -> bool:
        return self._is_pellet_shift_enabled

    @is_pellet_shift_enabled.setter
    def is_pellet_shift_enabled(self, value: bool) -> None:
        self._is_pellet_shift_enabled = value

    @property
    def is_auto_clamp_enabled(self) -> bool:
        return self._is_auto_clamp_enabled

    @is_auto_clamp_enabled.setter
    def is_auto_clamp_enabled(self, value: bool) -> None:
        self._is_auto_clamp_enabled = value

    @property
    def auto_clamp_no_activity_release_delay(self) -> float:
        return self._auto_clamp_no_activity_release_delay

    @auto_clamp_no_activity_release_delay.setter
    def auto_clamp_no_activity_release_delay(self, value: float) -> None:
        self._auto_clamp_no_activity_release_delay = value

    @property
    def auto_clamp_release_load_count(self) -> int:
        return self._auto_clamp_release_load_count

    @auto_clamp_release_load_count.setter
    def auto_clamp_release_load_count(self, value: int) -> None:
        self._auto_clamp_release_load_count = value

    @property
    def would_advance(self) -> bool:
        return self._would_advance

    @property
    def would_fallback(self) -> bool:
        return self._would_fallback

    def enter(self, context: PredicateContext, is_resume: bool = False) -> None:
        # If is_resume is True, the phase was in progress before this call.  Most likely, it was the active phase when
        # the system was stopped and serialized.  is_resume should not be called as True when started due to fallback
        # or advance, even if it is not the first time in the phase.
        if context.algorithm is not None:
            context.algorithm.reset_configuration()
            context.algorithm.pellet_delivery_enabled = self.is_pellet_delivery_enabled
            context.algorithm.pellet_cover_enabled = self.is_pellet_cover_enabled
            context.algorithm.baseline_intensity = self.starting_baseline_intensity
            context.algorithm.intersession_pellet_shift_enabled = self.is_pellet_shift_enabled

            if self.pellet_hands_min_distance > 0.0:
                context.algorithm.pellet_hands_min_distance = self.pellet_hands_min_distance
            else:
                # Need to explicitly "turn off" since the default might be to have it enabled.
                context.algorithm.pellet_hands_min_distance = math.inf

            context.algorithm.head_fixation_enabled = self.is_auto_clamp_enabled

            if self.auto_clamp_no_activity_release_delay > 0.0:
                context.algorithm.auto_clamp_no_activity_release_delay = self.auto_clamp_no_activity_release_delay

            if self._auto_clamp_release_load_count > 0:
                context.algorithm.auto_clamp_release_load_count = self._auto_clamp_release_load_count

        # TODO Not certain is_resume flag is properly set and considered for all transition types where this is called
        #  (advance vs. fallback vs. resume, etc) and between loaded existing progress vs. first time in protocol or
        #  phase.

        if not is_resume:
            # Do not need an init call if resuming.  It would overwrite any stored action progress.
            self._perform_session_actions(context, True)
            # Advance or fallback resets progress, except, we need to maintain the attempt count.
            self.progress = TrainingProgress(self.progress.phase_attempts + 1)

        self._evaluate_predicates(context)

    def exit(self, context: PredicateContext) -> None:
        pass

    def session_started(self) -> None:
        self.progress.session_count += 1
        self.progress.progress_resumed()

    def session_ended(self, context: PredicateContext) -> SessionResult:
        self.progress.progress_paused()

        self._perform_session_actions(context)

        return self._evaluate_predicates(context)

    def _evaluate_predicates(self, context: PredicateContext) -> SessionResult:
        self._would_advance = self._on_property_changed("would_advance", self._should_advance(context),
                                                        self._would_advance)

        self._would_fallback = self._on_property_changed("would_fallback", self._should_fallback(context),
                                                         self._would_fallback)

        if self._would_advance:
            return SessionResult.ADVANCE

        if self._would_fallback:
            return SessionResult.FALLBACK

        return SessionResult.NONE

    def _perform_session_actions(self, context: PredicateContext, is_init: bool = False) -> None:
        # Change delivery locations, magnet intensity, etc... as part of an individual training phase.

        # Actions care about whether it is an initialization call because some actions may be relative to a current
        # value.  For example, progressively move the pellet further way based on the position when the phase is first
        # entered.
        context.progress = self.progress
        for action in self.session_actions:
            action.evaluate(self, context, is_init)

    def _should_fallback(self, context: PredicateContext) -> bool:
        # Check time elapsed too long without enough successful reaches and revert to the last phase, etc.
        context.progress = self.progress
        if self.fallback_predicate is not None:
            return self.fallback_predicate.evaluate(self, context)

        return False

    def _should_advance(self, context: PredicateContext) -> bool:
        context.progress = self.progress
        # Move to the next training phase when any required conditions are met
        if self.advance_predicate is not None:
            return self.advance_predicate.evaluate(self, context) and self._are_actions_complete()

        return False

    def _are_actions_complete(self):
        return all(((not action.has_progress) or action.is_complete) for action in self.session_actions)

    def _on_property_changed(self, property_name, new_value, old_value):
        if old_value == new_value:
            return old_value

        for listener in self.property_changed:
            listener(property_name, new_value, old_value)

        return new_value

    def serialize_progress(self) -> Dict[str, Any]:
        progress = self.progress.to_dict()

        progress["action_context"] = [action.serialize_progress() for action in self.session_actions]

        return progress

    def deserialize_progress(self, progress_item: Dict[str, Any]) -> None:
        self.progress = TrainingProgress.from_dict(progress_item)

        if self.progress.action_context is not None:
            for idx, context in enumerate(self.progress.action_context):
                self.session_actions[idx].deserialize_progress(context)

    def to_dict(self) -> Dict[str, Any]:
        """Serialize to dictionary for JSON serialization"""
        data = {
            "phase_id": self.phase_id,
            "name": self.name,
            "description": self.description,

            "fallback_predicate": self.fallback_predicate.to_dict() if self.fallback_predicate else None,
            "advance_predicate": self.advance_predicate.to_dict() if self.advance_predicate else None,
            "session_actions": [action.to_dict() for action in self.session_actions],

            "is_pellet_delivery_enabled": self.is_pellet_delivery_enabled,
            "is_pellet_cover_enabled": self.is_pellet_cover_enabled,
            "starting_baseline_intensity": self.starting_baseline_intensity,
            "pellet_hands_min_distance": self.pellet_hands_min_distance,
            "is_pellet_shift_enabled": self.is_pellet_shift_enabled,
            "is_auto_clamp_enabled": self.is_auto_clamp_enabled,
            "auto_clamp_no_activity_release_delay": self.is_auto_clamp_enabled,
            "auto_clamp_release_load_count": self.is_auto_clamp_enabled
        }
        return humps.camelize(data)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Self:
        """Deserialize from dictionary"""
        data = humps.decamelize(data)
        phase = cls(data["phase_id"])
        phase.name = data["name"]
        phase.description = data["description"]

        phase.fallback_predicate = TrainingPredicate.from_dict(data.get("fallback_predicate", None))
        phase.advance_predicate = TrainingPredicate.from_dict(data.get("advance_predicate", None))
        if "session_actions" in data:
            phase.session_actions = [TrainingAction.from_dict(d) for d in data["session_actions"]]
        else:
            phase.session_actions = []

        phase.is_pellet_delivery_enabled = data.get("is_pellet_delivery_enabled", False)
        phase.is_pellet_cover_enabled = data.get("is_pellet_cover_enabled", False)
        phase.starting_baseline_intensity = data.get("starting_baseline_intensity", 5)
        phase.pellet_hands_min_distance = data.get("pellet_hands_min_distance", 0.0)
        phase.is_pellet_shift_enabled = data.get("is_pellet_shift_enabled", False)
        phase.is_auto_clamp_enabled = data.get("is_auto_clamp_enabled", False)
        phase.auto_clamp_no_activity_release_delay = data.get("auto_clamp_no_activity_release_delay", 2.0)
        phase.auto_clamp_release_load_count = data.get("auto_clamp_release_load_count", 5)

        return phase

    def status(self) -> str:
        status = ""

        return status

    def progress_status(self) -> str:
        return self.progress.status()



import importlib
from dataclasses import dataclass
from enum import IntEnum
from typing import Protocol, Optional, Dict, Any

import humps

from autotrainer.behavior.behavior_algorithm import BehaviorAlgorithm
from autotrainer.behavior.pellet_device_protocol import PelletDeviceProtocol
from autotrainer.behavior.tunnel_device_protocol import TunnelDeviceProtocol

from .training_progress import TrainingProgress


# Developer note: The use of protocols in this file name and contents is in reference to Python Protocols _not_ the
# user-facing/medical use of "protocol" in "training protocol".


class TrainingPhaseProtocol(Protocol):
    """
    Defined as a standalone protocol primarily to void circular imports between TrainingPhase and TrainingPredicate.
    """

    @property
    def progress(self) -> TrainingProgress: ...

    @property
    def starting_baseline_intensity(self) -> int: ...

    @property
    def is_pellet_cover_enabled(self) -> bool: ...

    @property
    def is_auto_clamp_enabled(self) -> bool: ...


class TrainingProgressState(IntEnum):
    Unknown = -1
    Initialized = 0
    Active = 10
    Paused = 20
    Failed = 30
    Complete = 40


@dataclass
class PredicateContext:
    progress: TrainingProgress = None
    algorithm: Optional[BehaviorAlgorithm] = None
    pellet_device: Optional[PelletDeviceProtocol] = None
    tunnel_device: Optional[TunnelDeviceProtocol] = None


class TrainingActionProtocol(Protocol):
    """
    Defined as a standalone protocol primarily to void circular imports between TrainingPhase and TrainingAction.
    """

    @property
    def has_progress(self) -> bool: ...

    @property
    def is_complete(self) -> bool: ...

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext, is_init: bool) -> bool: ...

    def serialize_progress(self) -> Dict[str, Any]: ...

    def deserialize_progress(self, data: Dict[str, Any]) -> None: ...

    def to_dict(self) -> Dict[str, Any]: ...

    @classmethod
    def from_dict(cls, data: Optional[Dict[str, Any]]): ...


def class_to_dict(name: str, module: str, props: Dict[str, Any]) -> Dict[str, Any]:
    data: Dict[str, Any] = {
        "type": name,
        "module": module
    }

    if props:
        data["properties"] = humps.camelize(props)

    return humps.camelize(data)


def class_from_dict(data: Optional[Dict[str, Any]]):
    if data is None:
        return None

    data = humps.decamelize(data)
    module = importlib.import_module(data["module"])
    class_ = getattr(module, data["type"])

    if "properties" in data and getattr(class_, "from_properties"):
        return class_.from_properties(humps.decamelize(data["properties"]))
    else:
        return class_()



from .training_protocols import PredicateContext, TrainingPhaseProtocol, TrainingProgressState, TrainingActionProtocol
from .training_progress import TrainingProgress
from .training_phase import TrainingPhase
from .training_plan import TrainingPlan

__all__ = ["TrainingProgress", "TrainingPhase", "PredicateContext", "TrainingPlan", "TrainingPhaseProtocol",
           "TrainingActionProtocol", "TrainingProgressState"]



#!/usr/bin/env python3
"""
Interactive console application for controlling TrainingSimulator objects.

This console provides command-line interface to load, save, and manipulate
training plans and progress data using the TrainingSimulator class.
"""

import json
from pathlib import Path
from typing import List

from autotrainer.training.training_simulator import TrainingSimulator


class TrainingConsole:
    """Interactive console for controlling TrainingSimulator instances."""

    def __init__(self):
        """Initialize the console with a new TrainingSimulator instance."""
        self.simulator = TrainingSimulator()
        self.last_command = ""
        self.running = True

    def run(self) -> None:
        """Run the main console loop."""
        print("Training Console - Interactive TrainingSimulator Control")
        print("Type 'help' or '?' for available commands, 'quit' to exit")
        print()

        while self.running:
            try:
                line = input(f"training> ").strip()

                # Allow repeating last command with empty input
                if not line and self.last_command:
                    line = self.last_command
                    print(f"(repeating: {line})")

                if line:
                    self.last_command = line
                    self.process_command(line)

            except (KeyboardInterrupt, EOFError):
                print("\nExiting...")
                break
            except Exception as error:
                print(f"Unexpected error: {error}")
                print("Type 'help' for available commands.")

    def process_command(self, command_line: str) -> None:
        """Parse and execute a command."""
        parts = command_line.split()
        if not parts:
            return

        command = parts[0].lower()
        args = parts[1:]

        # Command dispatch
        command_map = {
            "help": self.cmd_help,
            "?": self.cmd_help,
            "quit": self.cmd_quit,
            "exit": self.cmd_quit,
            "status": self.cmd_status,
            "load": self.cmd_load_plan,
            "save": self.cmd_save_plan,
            "load_progress": self.cmd_load_progress,
            "save_progress": self.cmd_save_progress,
            "reset_progress": self.cmd_reset_progress,
            "start_session": self.cmd_start_session,
            "end_session": self.cmd_end_session,
            "inc_reaches": self.cmd_inc_reaches,
            "inc_presented": self.cmd_inc_presented,
            "inc_consumed": self.cmd_inc_consumed,
            "info": self.cmd_info,
            "progress": self.cmd_progress,
        }

        if command in command_map:
            command_map[command](args)
        else:
            print(f"Unknown command: {command}. Type 'help' for available commands.")

    def cmd_help(self, args: List[str]) -> None:
        """Display help information."""
        print("Available commands:")
        print()
        print("File Operations:")
        print("  load_plan <filepath>       Load a training plan from JSON file")
        print("  save_plan [filepath]       Save current training plan to JSON file")
        print("  load_progress <filepath>   Load training progress from JSON file")
        print("  save_progress [filepath]   Save current training progress to JSON file")
        print()
        print("Training Control:")
        print("  reset_progress             Reset all progress data in current plan")
        print("  inc_reaches [quantity]     Increase successful reaches count (default: 1)")
        print("  inc_presented [quantity]   Increase pellets presented count (default: 1)")
        print("  inc_consumed [quantity]    Increase pellets consumed count (default: 1)")
        print()
        print("Information:")
        print("  status                     Show simulator status")
        print("  info                       Display current training plan details")
        print("  progress                   Display current progress statistics")
        print()
        print("Utility:")
        print("  help, ?                    Show this help message")
        print("  quit, exit                 Exit the console")
        print()
        print("Notes:")
        print("  - Press Enter without typing to repeat the last command")
        print("  - File paths can be relative or absolute")

    def cmd_quit(self, args: List[str]) -> None:
        """Exit the console."""
        self.running = False

    def cmd_load_plan(self, args: List[str]) -> None:
        """Load a training plan from a JSON file."""
        if not args:
            print("Error: load_plan requires a file path")
            print("Usage: load_plan <filepath>")
            return

        try:
            file_path = Path(args[0]).resolve()
            self.simulator.load_training_plan(file_path)
            print(f"Successfully loaded training plan from: {file_path}")
            if self.simulator.training_plan:
                print(f"Plan name: {self.simulator.training_plan.name or 'Unnamed'}")
                print(f"Number of phases: {len(self.simulator.training_plan.phases)}")
        except FileNotFoundError:
            print(f"Error: File not found: {args[0]}")
        except PermissionError:
            print(f"Error: Permission denied accessing: {args[0]}")
        except json.JSONDecodeError as error:
            print(f"Error: Invalid JSON in file: {error}")
        except Exception as error:
            print(f"Error loading training plan: {error}")

    def cmd_save_plan(self, args: List[str]) -> None:
        """Save the current training plan to a JSON file."""
        try:
            file_path = Path(args[0]).expanduser().resolve() if args else None
            self.simulator.save_training_plan(file_path)
            target_path = file_path or self.simulator.plan_file_path
            print(f"Successfully saved training plan to: {target_path}")
        except ValueError as error:
            print(f"Error: {error}")
            if not args:
                print("Usage: save_plan <filepath>")
        except PermissionError:
            file_arg = args[0] if args else "current path"
            print(f"Error: Permission denied writing to: {file_arg}")
        except Exception as error:
            print(f"Error saving training plan: {error}")

    def cmd_load_progress(self, args: List[str]) -> None:
        """Load training progress from a JSON file."""
        if not args:
            print("Error: load_progress requires a file path")
            print("Usage: load_progress <filepath>")
            return

        try:
            file_path = Path(args[0]).expanduser().resolve()
            self.simulator.load_training_progress(file_path)
            print(f"Successfully loaded training progress from: {file_path}")
        except ValueError as error:
            print(f"Error: {error}")
        except FileNotFoundError:
            print(f"Error: File not found: {args[0]}")
        except PermissionError:
            print(f"Error: Permission denied accessing: {args[0]}")
        except json.JSONDecodeError as error:
            print(f"Error: Invalid JSON in file: {error}")
        except Exception as error:
            print(f"Error loading training progress: {error}")

    def cmd_save_progress(self, args: List[str]) -> None:
        """Save the current training progress to a JSON file."""
        try:
            file_path = Path(args[0]).expanduser().resolve() if args else None
            self.simulator.save_training_progress(file_path)
            target_path = file_path or self.simulator.progress_file_path
            print(f"Successfully saved training progress to: {target_path}")
        except ValueError as error:
            print(f"Error: {error}")
            if not args:
                print("Usage: save_progress <filepath>")
        except PermissionError:
            file_arg = args[0] if args else "current path"
            print(f"Error: Permission denied writing to: {file_arg}")
        except Exception as error:
            print(f"Error saving training progress: {error}")

    def cmd_reset_progress(self, args: List[str]) -> None:
        """Reset all progress data in the current training plan."""
        if not self.simulator.training_plan:
            print("Error: No training plan loaded. Load a plan first.")
            return

        # Ask for confirmation
        try:
            response = input("Are you sure you want to reset all progress data? (y/N): ").strip().lower()
            if response in ["y", "yes"]:
                self.simulator.reset_progress()
                print("Progress data has been reset for all phases.")
            else:
                print("Reset cancelled.")
        except (KeyboardInterrupt, EOFError):
            print("\nReset cancelled.")

    def cmd_start_session(self, args: List[str]) -> None:
        """Start behavior session."""
        if not self.simulator.training_plan:
            print("Error: No training plan loaded. Load a plan first.")
            return
        self.simulator.start_session()

    def cmd_end_session(self, args: List[str]) -> None:
        """End behavior session."""
        if not self.simulator.training_plan:
            print("Error: No training plan loaded. Load a plan first.")
            return
        self.simulator.end_session()

    def cmd_inc_reaches(self, args: List[str]) -> None:
        """Increase successful reaches count."""
        if not self.simulator.training_plan:
            print("Error: No training plan loaded. Load a plan first.")
            return

        quantity = 1
        if args:
            try:
                quantity = int(args[0])
                if quantity <= 0:
                    print("Error: Quantity must be a positive integer")
                    return
            except ValueError:
                print("Error: Invalid quantity. Must be a positive integer.")
                return

        try:
            self.simulator.increase_successful_reaches(quantity)
            print(f"Increased successful reaches by {quantity}")
        except Exception as error:
            print(f"Error: {error}")

    def cmd_inc_presented(self, args: List[str]) -> None:
        """Increase pellets presented count."""
        if not self.simulator.training_plan:
            print("Error: No training plan loaded. Load a plan first.")
            return

        quantity = 1
        if args:
            try:
                quantity = int(args[0])
                if quantity <= 0:
                    print("Error: Quantity must be a positive integer")
                    return
            except ValueError:
                print("Error: Invalid quantity. Must be a positive integer.")
                return

        try:
            self.simulator.increase_pellets_presented(quantity)
            print(f"Increased pellets presented by {quantity}")
        except Exception as error:
            print(f"Error: {error}")

    def cmd_inc_consumed(self, args: List[str]) -> None:
        """Increase pellets consumed count."""
        if not self.simulator.training_plan:
            print("Error: No training plan loaded. Load a plan first.")
            return

        quantity = 1
        if args:
            try:
                quantity = int(args[0])
                if quantity <= 0:
                    print("Error: Quantity must be a positive integer")
                    return
            except ValueError:
                print("Error: Invalid quantity. Must be a positive integer.")
                return

        try:
            self.simulator.increase_pellets_consumed(quantity)
            print(f"Increased pellets consumed by {quantity}")
        except Exception as error:
            print(f"Error: {error}")

    def cmd_status(self, args: List[str]) -> None:
        """Display current simulator status."""
        print("---------------------------------\n")
        print(self.simulator.status())
        print("---------------------------------\n")

    def cmd_info(self, args: List[str]) -> None:
        """Display detailed information about the current training plan."""
        print("---------------------------------\n")
        print(self.simulator.info())
        print("---------------------------------\n")

    def cmd_progress(self, args: List[str]) -> None:
        """Display current progress statistics."""
        print("---------------------------------\n")
        print(self.simulator.progress_status())
        print("---------------------------------\n")


def main():
    """Main entry point for the training console application."""
    console = TrainingConsole()
    console.run()


if __name__ == "__main__":
    main()
